/*
 * Segment 2 (KEx),  SWE2.4 Email/Fax/SMS
 * Copyright (C) 2007 BitCtrl Systems GmbH
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
 * details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA.
 *
 * Contact Information:
 * BitCtrl Systems GmbH
 * Weißenfelser Straße 67
 * 04229 Leipzig
 * Phone: +49 341-490670
 * mailto: info@bitctrl.de
 */

package de.bsvrz.kex.emailfaxsms;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import de.bsvrz.sys.funclib.bitctrl.daf.LogTools;
import de.bsvrz.sys.funclib.debug.Debug;
import gnu.io.SerialPort;

/**
 * Schreibt asynchron Daten aus dem Puffer auf die seriellen Schnittstelle.
 *
 * @author BitCtrl Systems GmbH, Bachraty
 */
public class AsyncSerialPortWriter extends Thread {

	/**
	 * Der Logger.
	 */
	protected static final Debug LOGGER = Debug.getLogger();

	/**
	 * Ausgabestream serielle Schnittstelle.
	 */
	protected OutputStream os;

	/**
	 * Eingabestream serielle Schnittstelle (f&uuml;r die
	 * SW-Datenflusssteuerung).
	 */
	protected InputStream is;

	/**
	 * Parameter fuer Debug-Zwecke.
	 */
	protected boolean mitDebug = false;

	/**
	 * Anzahl der zus&auml;tzlich eingef&uuml;gten DLE's.
	 */
	protected int extraDLE = 0;

	/**
	 * AT-Kommando zum versenden.
	 */
	protected String atKommando = null;

	/**
	 * Bytes zum versenden.
	 */
	protected byte[] bytePuffer = null;
	/**
	 * Groesse des Byte-Strings zu versenden.
	 */
	protected int bytePufferDatenGroesse = 0;

	/**
	 * Flag, ob beendet werden soll.
	 */
	protected boolean beenden = false;
	/**
	 * Anzahl der versendeten Bytes im letzten Chunk.
	 */
	protected int versendeteBytes = 0;
	/**
	 * Anzahl der versendeten Bytes - Summe.
	 */
	protected int versendeteBytesTotal = 0;
	/**
	 * Die serielle Schnittstelle.
	 */
	protected SerialPort port;
	/**
	 * Der Reader-Thread.
	 */
	protected AsyncSerialPortReader reader;

	/**
	 * Flag, ob die SW-DatenFlusssteuerung eingeschaltet ist.
	 */
	protected boolean swDatenFlussSteuerung = true;
	/**
	 * Bestimmt ob Daten versendet werden k&ouml;nnen, wenn die
	 * SW-DatenFlusssteuerung eingeschaltet ist.
	 */
	protected boolean datenFlussGeeignet = false;

	/**
	 * Wird nur als Mutex benutzt.
	 */
	protected Object mutex = new Object();

	/**
	 * Ready-Flag.
	 */
	private boolean ready = false;

	/**
	 * Standardkonstruktor.
	 *
	 * @param port
	 *            Die serielle Schnittstelle
	 * @param pufferSize
	 *            PufferGroesse
	 * @param read
	 *            Der Reader Thread
	 * @param swFlowControl
	 *            benutze Software-Flusssteuerung ja/nein
	 * @throws IOException
	 *             bei Ausnahmen
	 */
	public AsyncSerialPortWriter(final SerialPort port, final int pufferSize, final AsyncSerialPortReader read,
			final boolean swFlowControl) throws IOException {
		this.os = port.getOutputStream();
		this.is = port.getInputStream();
		this.bytePuffer = new byte[pufferSize];
		this.swDatenFlussSteuerung = swFlowControl;
		this.reader = read;
		this.port = port;
	}

	/**
	 * Beendet den Versand-Thread.
	 */
	public synchronized void beenden() {
		this.beenden = true;
		try {
			this.os.close();
		} catch (final IOException e) {
			LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningDatenSchreiben, e.getMessage());
		}
		this.notify();
	}

	/**
	 * Wird am Ende des bin&auml;ren Modus aufgerufen, wenn OK als Antwort vom
	 * Modem gekommen ist.
	 */
	public void endBinaryModus() {
		setSwFlowStatus(false);
		reader.setStopped(false);
		if (mitDebug) {
			if (versendeteBytes > 0) {
				System.out.println("## SENT: " + versendeteBytes + " Bytes");
			}
		}
		versendeteBytesTotal += versendeteBytes;
		if (mitDebug) {
			System.out.println("\n### TOTAL BYTES SENT : " + versendeteBytesTotal);
			System.out.println("\n### EXTRA DLEs SENT : " + extraDLE);
		}
	}

	/**
	 * Wird beim Event OutputBufferEmpty aufgerufen.
	 */
	public void outputPufferLeer() {
		synchronized (mutex) {
			this.mutex.notify();
		}
	}

	/**
	 * Gibt zurück, ob der Thread bereit ist.
	 *
	 * @return bereit oder nicht
	 */
	public boolean ready() {
		return ready;
	}

	/**
	 * Kehrt die Reihenfolge der Bits im Byte um.
	 *
	 * @param b
	 *            Byte
	 * @return Byte mit umgekehrten Bits
	 */
	private byte reverseBits(final byte b) {
		byte c = 0;

		if ((b & 0x01) != 0) {
			c |= 0x80;
		}
		if ((b & 0x02) != 0) {
			c |= 0x40;
		}
		if ((b & 0x04) != 0) {
			c |= 0x20;
		}
		if ((b & 0x08) != 0) {
			c |= 0x10;
		}
		if ((b & 0x10) != 0) {
			c |= 0x08;
		}
		if ((b & 0x20) != 0) {
			c |= 0x04;
		}
		if ((b & 0x40) != 0) {
			c |= 0x02;
		}
		if ((b & 0x80) != 0) {
			c |= 0x01;
		}

		return (c);
	}

	/**
	 * Der Arbeiter-Thread.
	 */
	@Override
	public synchronized void run() {
		while (!beenden) {
			try {
				// Nur im binaeren modus gibt es Datenflusssteuerung
				// im Kommando-Modus schreibt gleich in die Schnittstelle
				if (atKommando != null) {
					LogTools.log(Debug.getLogger(), EmailFaxSmsMsg.FinestModemCommand, atKommando);

					os.write(atKommando.getBytes());
					os.write(AsyncSerialPort.ENDL);
					os.flush();
					atKommando = null;
					this.notify();
				} else if (bytePufferDatenGroesse > 0) {
					LogTools.log(Debug.getLogger(), EmailFaxSmsMsg.FineModemWriteBytes, bytePufferDatenGroesse);
					for (int i = 0; i < bytePufferDatenGroesse; i++) {
						writeByte(bytePuffer[i]);
					}
					bytePufferDatenGroesse = 0;
					os.flush();
					this.notify();
				}

				ready = true;

				if (!beenden) {
					this.wait();
				}
			} catch (final InterruptedException e) {
				LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningThreadUnterbrochen);
			} catch (final IOException e) {
				LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler, e.getMessage());
			}
		}
		LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.FinerThreadBeendet, this.getClass().getName());
	}

	/**
	 * Sendet ein Byte.
	 *
	 * @param timeout
	 *            timeout
	 * @param b
	 *            Byte
	 * @return true, wenn byte erfolgreich versendet wurde
	 */
	public synchronized boolean sendByte(final long timeout, final byte b) {
		this.bytePuffer[0] = b;
		this.bytePufferDatenGroesse = 1;
		this.notify();
		try {
			this.wait(timeout);
			if (bytePufferDatenGroesse == 0) {
				return true;
			}
		} catch (final InterruptedException e) {
			LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler);
		}
		bytePufferDatenGroesse = 0;
		return false;
	}

	/**
	 * Sendet eine Byte-Kette ohne etwas zu ersetzen.
	 *
	 * @param timeout
	 *            Timeout
	 * @param cmd
	 *            Byte-Kommando zu versenden
	 * @param off
	 *            Offset
	 * @param len
	 *            Length
	 * @return true, wenn Byte-Kette erfolgreich versendet wurde
	 */
	public synchronized boolean sendBytes(final long timeout, final byte[] cmd, final int off, final int len) {
		if (cmd.length > bytePuffer.length) {
			return false;
		}
		System.arraycopy(cmd, off, this.bytePuffer, 0, len);
		this.bytePufferDatenGroesse = cmd.length;
		this.notify();
		try {
			this.wait(timeout);
			if (bytePufferDatenGroesse == 0) {
				return true;
			}
		} catch (final InterruptedException e) {
			LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler);
		}
		bytePufferDatenGroesse = 0;
		return false;
	}

	/**
	 * Sendet ein AT Kommando.
	 *
	 * @param timeout
	 *            Timeout
	 * @param cmd
	 *            Kommando
	 * @return true, wenn erfolgreich versendet
	 */
	public synchronized boolean sendCmd(final long timeout, final String cmd) {
		this.atKommando = cmd;
		this.notify();
		try {
			this.wait(timeout);
			if (cmd == null) {
				return true;
			}
		} catch (final InterruptedException e) {
			LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningDatenVersandUnterbrochen);
		}
		// cmd = null;
		return false;
	}

	/**
	 * Versendet Daten im bin&auml;ren Modus. Jedes <code>DLE</code> wird
	 * gedoppelt, damit es vom Modem korrekt interpertiert wird
	 *
	 * @param timeout
	 *            Timeout
	 * @param data
	 *            Daten
	 * @param off
	 *            Offset
	 * @param len
	 *            Laenge
	 * @param reverseBitOrder
	 *            Flag, ob die Bitorder umgekehrt werden soll
	 * @return Anzahl der versendeten Bytes
	 * @throws Exception
	 *             bei Ausnahmen
	 */

	public synchronized int sendData(final long timeout, final byte[] data, final int off, final int len,
			final boolean reverseBitOrder) throws Exception {
		long start;
		int versendet = 0;
		int i = 0, j = off, extradle = 0;
		long resttime = timeout;

		if (mitDebug) {
			System.out.print("\n######## Bytes TO send:" + len + "\nSEND:");
		}
		while (versendet < len) {
			// Teilt die Daten auf Teile wie die Puffergroesse und
			// ersetzt DLEs durch Doppel-DLEs

			for (extradle = i = 0; (i < (bytePuffer.length - 1)) && (j < (len + off)); i++, j++) {
				byte wb = data[j];
				if (reverseBitOrder) {
					wb = reverseBits(wb);
				}

				if (wb == AsyncSerialPort.DLE) {
					this.bytePuffer[i++] = AsyncSerialPort.DLE;
					extradle++;
					extraDLE++;
				}
				bytePuffer[i] = wb;
			}
			this.bytePufferDatenGroesse = i;
			this.notify();
			try {
				start = System.currentTimeMillis();
				this.wait(resttime);
				resttime -= System.currentTimeMillis() - start;
			} catch (final InterruptedException e) {
				LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler);
			}
			if (bytePufferDatenGroesse == 0) {
				versendet += i - extradle;
			} else {
				return versendet;
			}
			if (resttime < 0) {
				return versendet;
			}
		}
		if (mitDebug) {
			if (versendeteBytes > 0) {
				System.out.println("\n## SENT: " + versendeteBytes + " Bytes");
			}
		}
		versendeteBytesTotal += versendeteBytes;
		versendeteBytes = 0;
		bytePufferDatenGroesse = 0;
		return versendet;
	}

	/**
	 * Setzt den Debug-Parameter.
	 *
	 * @param debug
	 *            true, wenn Debug Ein-ist
	 */
	public void setDebug(final boolean debug) {
		this.mitDebug = debug;
	}

	/**
	 * Setzt die SwDatenFlusssteuerung.
	 *
	 * @param flussGeeignet
	 *            <code>true</code>, wenn SwDatenFlusssteuerung benutzt werden
	 *            soll
	 */
	public void setSwFlowStatus(final boolean flussGeeignet) {
		this.datenFlussGeeignet = flussGeeignet;
	}

	/**
	 * Wird bei Start des bin&auml;ren Modus aufgerufen, wenn CONNECT als
	 * Antwort vom Modem gekommen ist.
	 */
	public void startBinaryModus() {
		setSwFlowStatus(true);
		reader.setStopped(true);
	}

	/**
	 * Schreibt ein Byte auf die serielle Schnittstelle.
	 *
	 * @param b
	 *            Byte zum Versenden
	 */
	protected synchronized void writeByte(final byte b) {
		if (swDatenFlussSteuerung) {
			synchronized (mutex) {
				if (!datenFlussGeeignet) {
					if (mitDebug) {
						if (versendeteBytes > 0) {
							System.out.println("## SENT: " + versendeteBytes + " Bytes");
						}
						System.out.println("SW Flow : OFF, wait");
						versendeteBytesTotal += versendeteBytes;
						versendeteBytes = 0;
					}

					if (mitDebug) {
						System.out.print("SW Flow :" + ((datenFlussGeeignet) ? "ON" : "OFF") + "\nSEND:");
					}
				}
				while (!datenFlussGeeignet) {
					try {
						if (is.available() > 0) {
							final int x = is.read();

							if (x == AsyncSerialPort.XON) {
								datenFlussGeeignet = true;
								System.out.println("WRITER:<XON>");
							} else if (x == AsyncSerialPort.XOFF) {
								datenFlussGeeignet = false;
								System.out.println("WRITER:<XOFF>");
							}
						}
						Thread.sleep(1);
					} catch (final InterruptedException e) {
						LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningThreadUnterbrochen);
					} catch (final IOException e) {
						LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler, e.getMessage());
					}
				}
				try {
					os.write(b);
					versendeteBytes++;
					if (mitDebug && (versendeteBytes < 20)) {
						System.out.print(String.format(" 0x%xh", b));
					}
				} catch (final IOException e) {
					LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler);
				}
			}
		} else {
			try {
				os.write(b);
				versendeteBytes++;
				if (mitDebug && (versendeteBytes < 20)) {
					System.out.print(String.format(" 0x%xh", b));
				}
			} catch (final IOException e) {
				LogTools.log(AsyncSerialPortWriter.LOGGER, EmailFaxSmsMsg.WarningSchreibFehler, e.getMessage());
			}
		}
	}

}
